/**
 * Runtime representation of an entity value.
 *
 */
export declare class Entity {
    value: string;
    display: string | null;
    /**
     * Construct a new entity value.
     *
     * @param {string} value - the entity identifier value
     * @param {string|null} [display] - optional human-readable display name for the entity
     */
    constructor(id: string, display?: string | null);
    toString(): string;
    /**
     * Compile the entity value to JS code.
     *
     * @return {string} JS code that evaluates to this entity value
     * @package
     */
    toJSSource(): string;
    /**
     * Check if this JS value looks like an entity.
     *
     * @param {any} obj - the JS value to check
     * @return {boolean} `true` if the value is a string or an instance of this class
     * @package
     */
    static isEntity(obj: unknown): obj is (Entity | string);
}
export interface LocationLike {
    x: number;
    y: number;
    display?: string | null;
}
/**
 * Runtime representation of a location value.
 *
 */
export declare class Location implements LocationLike {
    x: number;
    y: number;
    display: string | null;
    /**
     * Construct a new location value.
     *
     * @param {number} lat - latitude in degrees (between -90 and 90)
     * @param {number} lon - longitude in degrees (between -180 and 180)
     * @param {string|null} [display] - human-readable display name for this location
     */
    constructor(lat: number, lon: number, display?: string | null);
    /**
     * Latitude value.
     */
    get lat(): number;
    /**
     * Longitude value.
     */
    get lon(): number;
    toString(): string;
    /**
     * Compile the location value to JS code.
     *
     * @return {string} JS code that evaluates to this location value
     * @package
     */
    toJSSource(): string;
    /**
     * Check if this JS value looks like an location.
     *
     * For compatibility reasons, the runtime representation of a location is not required
     * to be an instance of the {@link Builtin.Location} class, and can be any object
     * with `x` and `y` own properties.
     *
     * @param {any} obj - the JS value to check
     * @return {boolean} `true` if the value looks like a location
     * @package
     */
    static isLocation(obj: unknown): obj is LocationLike;
}
/**
 * Runtime representation of a time value.
 *
 */
export declare class Time {
    hour: number;
    minute: number;
    second: number;
    /**
     * Construct a new time value.
     *
     * @param {number} hour - hour value (between 0 and 11)
     * @param {number} minute - minute value (between 0 and 59)
     * @param {number} [second] - second value (between 0 and 59)
     */
    constructor(hour: number, minute: number, second?: number);
    static fromSeconds(seconds: number): Time;
    /**
     * Convert this time value to the number of seconds since midnight.
     *
     * This can be used to compare time values using `<` and `>`
     * @return {number} the number of seconds since midnight
     */
    valueOf(): number;
    toString(): string;
    toJSON(): string;
    /**
     * Compile the time value to JS code.
     *
     * @return {string} JS code that evaluates to this time value
     * @package
     */
    toJSSource(): string;
}
/**
 * Runtime representation of a currency value.
 *
 */
export declare class Currency {
    value: number;
    code: string;
    constructor(value: number, code: string);
    valueOf(): number;
    toString(): string;
    toLocaleString(locale: string, options?: Intl.NumberFormatOptions): string;
    toJSSource(): string;
}
export interface RecurrentTimeRuleLike {
    beginTime: Time;
    endTime: Time;
    interval: number;
    frequency: number;
    dayOfWeek: number | null;
    beginDate: Date | null;
    endDate: Date | null;
    subtract: boolean;
}
/**
 * Runtime representation of a recurrent time rule (a single item in a recurrent time specification).
 *
 */
export declare class RecurrentTimeRule implements RecurrentTimeRuleLike {
    beginTime: Time;
    endTime: Time;
    interval: number;
    frequency: number;
    dayOfWeek: number | null;
    beginDate: Date | null;
    endDate: Date | null;
    subtract: boolean;
    constructor({ beginTime, endTime, interval, frequency, dayOfWeek, beginDate, endDate, subtract }: RecurrentTimeRuleLike);
    toString(): string;
    toJSSource(): string;
    contains(dateOrTime: Date | Time): boolean;
}
export declare class Aggregation {
    type: string;
    field: string;
    cols: string;
    count: number;
    constructor(type: string, field: string, cols: string, count: number);
}
